/**
 * Copyright (c) 2001-2006 eXo Platform SARL, http://www.exoplatform.com/
 *
 * All rights reserved. Please look at license.txt in info directory for
 * more license detail.
 *
 * -------------------------------------------------------------------------
 * $Id: ExoSignerApplet.java 10718 2006-11-29 02:54:54Z rmarins $
 * -------------------------------------------------------------------------
 */
package org.exoplatform.applications.signerclient;

import java.awt.Dimension;
import java.io.File;

import javax.swing.JApplet;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import org.exoplatform.applet.Invokable;
import org.exoplatform.applet.logging.Logger;
import org.exoplatform.applet.logging.file.FileLoggerFactory;
import org.exoplatform.applet.plugin.AppletPlugin;
import org.exoplatform.applet.plugin.Pluggable;
import org.exoplatform.applet.scripting.ScriptableBrowserSupport;
import org.exoplatform.applet.scripting.ScriptingBrowser;
import org.exoplatform.applet.scripting.ScriptingObject;
import org.exoplatform.applications.signerclient.crypto.CryptoException;
import org.exoplatform.applications.signerclient.crypto.CryptoPlugin;
import org.exoplatform.applications.signerclient.util.SignerUtil;

/**
 * @version $Rev: 10718 $ $Date: 2006-11-29 09:54:54 +0700 (Wed, 29 Nov 2006) $
 * @author <a href="mailto:rafael.marins@exoplatform.com">Rafael Marins</a>
 */
public class ExoSignerApplet extends JApplet implements Pluggable, Invokable,
        ScriptableBrowserSupport {

	/**
	 * Serialization object version uid.
	 */
	private static final long serialVersionUID = 1L;

	private static final String LOG_SOURCE_CLASS = "ExoSignerApplet";

	private static final String DEFAULT_ALGORITHM = "SHA1withRSA";

	private static final String DEFAULT_CERTIFICATE_TYPE = "X.509";

	private static final String FALSE = Boolean.FALSE.toString();

	private static final String TRUE = Boolean.TRUE.toString();

	private Logger logger = FileLoggerFactory.getInstance().getLogger(
	        ExoSignerApplet.class.getName());

	private CryptoPlugin cryptoPlugin;

	private ScriptingBrowser browser;

	private SignerParameters parameters;

	private SignStatus status;

	/**
	 * Exo Signer Applet constructor.
	 */
	public ExoSignerApplet() {
		super();
	}

	public void init() {

		final String log_method = "init".intern();

		logger.entering(LOG_SOURCE_CLASS, log_method);

		SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                initializeGUI();
            }
        });

		super.init();

		logger.exiting(LOG_SOURCE_CLASS, log_method);
	}

	public void initializeGUI() {
		
		this.setSize(new Dimension(222, 155));

		status = new SignStatus(222, 155);
		status.showSigning();

		JPanel contentPane = new JPanel();
		contentPane.setLayout(null);
		contentPane.add(status.getShowingLabel(), null);

		this.add(contentPane);
		this.repaint();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.exoplatform.applet.scripting.ScriptableBrowserSupport#setBrowser(org.exoplatform.applet.scripting.ScriptingBrowser)
	 */
	public void setBrowser(ScriptingBrowser browser) {
		this.browser = browser;
	}

	private SignerParameters getParameters() {
		if (parameters == null)
			parameters = new SignerParameters(this);
		return parameters;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.exoplatform.applet.plugin.Pluggable#addPlugin(org.exoplatform.applet.plugin.AppletPlugin)
	 */
	public void addPlugin(AppletPlugin plugin) {
		if (plugin instanceof CryptoPlugin)
			this.cryptoPlugin = (CryptoPlugin) plugin;
	}

	private void showMessage(String msg) {
		browser.getWindow().call("userMessage", new Object[] {msg});
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.exoplatform.applet.Invokable#invoke(java.lang.String,
	 *      java.lang.String[])
	 */
	public Object invoke(String methodName, String[] args) {
		Object result = null;

		if ("signInputDigest".equals(methodName))
			result = signInputDigest(args[0], args[1]);
		else if ("showMessage".equals(methodName))
			showMessage(args[0]);

		return result;
	}

	public String signInputDigest(String keyStoreFile, String keyStorePassword) {

		final String log_method = "signInputDigest".intern();

		logger.entering(LOG_SOURCE_CLASS, log_method);

		SignerParameters params = getParameters();

		String formName = params.getSigningFormNameParam();
		String encodedDigestFieldName = params.getEncodedDigestFieldNameParam();

		// retrieve parameters values from the HTML form
		String encodedDigest = getFormField(formName, encodedDigestFieldName);
		byte[] hash = SignerUtil.base64Decode(encodedDigest);

		SignatureInfo sinfo = null;

		try {
			String algorithm = getAlgorithm();

			// sign and set result values in the HTML form
			sinfo = sign(hash, algorithm, keyStoreFile, keyStorePassword);

		} catch (ExoSignerException e) {
			logger.throwing(LOG_SOURCE_CLASS, log_method, e);
			doFailure(e.getMessage());
			return FALSE;
		}

		// set results into HTML form fields
		String signatureFieldName = params.getSignatureFieldNameParam();
		String certChainFieldName = params.getCerticateChainFieldNameParam();

		String encodedSignature = SignerUtil.base64Encode(sinfo.getSignature());
		String encodedCertPath = SignerUtil.base64Encode(sinfo.getCertPath());

		setFormField(formName, signatureFieldName, encodedSignature);
		setFormField(formName, certChainFieldName, encodedCertPath);

		doSuccess();

		return TRUE;
	}

	/**
     * @param string
     */
    private void doSuccess() {
	    status.showSigning();
//	    showMessage(message);
    }

	/**
     * @param message
     */
    private void doFailure(String message) {
    	status.showFailure();
    	showMessage(message);
    }

	private SignatureInfo sign(byte[] hash, String algorithm,
	        String keyStoreFile, String keyStorePassword)
	        throws ExoSignerException {

		SignatureInfo signInfo = new SignatureInfo();

		// load user keystore
		Object userKeyStore = loadUserKeyStore(keyStoreFile, keyStorePassword);

		// extract certificate chain from user's keystore
		byte[] certPath = getCertPath(userKeyStore);
		signInfo.setCertPath(certPath);

		// retrieve private key from the user's keystore
		Object signingKey = getPrivateKey(userKeyStore, keyStorePassword);

		// perform signing
		byte[] signature = doSign(hash, algorithm, signingKey);
		signInfo.setSignature(signature);

		return signInfo;
	}

	/**
	 * @param hash
	 * @param digestAlgo
	 * @param encryptAlgo
	 * @param privateKey
	 * @return
	 * @throws ExoSignerException
	 */
	private byte[] doSign(byte[] hash, String algorithm, Object privateKey)
	        throws ExoSignerException {

		byte[] signature = null;

		try {
			signature = cryptoPlugin.sign(hash, algorithm, privateKey);
		} catch (CryptoException e) {
			throw new ExoSignerException("Signing failed. " + e.getMessage(), e);
		}

		return signature;
	}

	/**
	 * @return
	 */
	private String getAlgorithm() {

		SignerParameters params = getParameters();

		String algo = params.getAlgorithmParam();

		// TODO if algorithm is unknown use default value also

		if (algo == null)
			algo = DEFAULT_ALGORITHM;

		return algo;
	}

	/**
	 * @param userKeyStore
	 * @return
	 * @throws ExoSignerException
	 */
	private byte[] getCertPath(Object keyStore) throws ExoSignerException {

		byte[] certPath = null;

		try {
			Object[] chain = cryptoPlugin.getCertificationChain(keyStore);

			if (chain == null)
				throw new ExoSignerException(
				        "Unable to extract certificate chain from user certificate KeyStore.");

			certPath = cryptoPlugin
			        .getCertPath(chain, DEFAULT_CERTIFICATE_TYPE);

		} catch (CryptoException e) {
			throw new ExoSignerException(
			        "Unable to extract certificate chain from user certificate KeyStore.",
			        e);
		}

		return certPath;
	}

	/**
	 * @param keyStore
	 * @return
	 * @throws ExoSignerException
	 */
	private Object getPrivateKey(Object keyStore, String password)
	        throws ExoSignerException {

		Object key = null;

		try {
			key = cryptoPlugin.getPrivateKey(keyStore, password);
		} catch (CryptoException e) {
			throw new ExoSignerException("Failed to load Private Key.", e);
		}

		return key;
	}

	/**
	 * @return
	 * @throws ExoSignerException
	 */
	private Object loadUserKeyStore(String path, String password)
	        throws ExoSignerException {

		Object keyStore = null;

		File file = new File(path);

		// TODO Check if keystore file exists

		try {
			// TODO Include keystore type mapping
			keyStore = cryptoPlugin.loadKeyStore(file, password, "PKCS12");
		} catch (CryptoException e) {
			throw new ExoSignerException(
			        "Failed to load user certificate KeyStore.", e);
		}

		return keyStore;
	}

	private String getFormField(String formName, String fieldName) {
		ScriptingObject form = browser.getForm(formName);
		ScriptingObject field = (ScriptingObject) form.getMember(fieldName);
		String value = browser.getMemberValue(field);
		return value;
	}

	private void setFormField(String formName, String fieldName, Object value) {
		ScriptingObject form = browser.getForm(formName);
		ScriptingObject field = (ScriptingObject) form.getMember(fieldName);
		browser.setMemberValue(field, value);
	}

}