/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import java.util.ArrayList;
import java.util.List;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.access.AccessControlPolicy;
import org.exoplatform.services.jcr.config.RepositoryEntry;
import org.exoplatform.services.jcr.config.WorkspaceEntry;
import org.exoplatform.services.jcr.core.ExtendedPropertyType;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeTypeManager;
import org.exoplatform.services.jcr.datamodel.IllegalPathException;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeDataPersister;
import org.exoplatform.services.jcr.impl.dataflow.DataManager;
import org.exoplatform.services.jcr.impl.dataflow.MutableItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.TransientNodeData;
import org.exoplatform.services.jcr.impl.dataflow.TransientPropertyData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.jcr.impl.util.EntityCollection;
import org.exoplatform.services.jcr.util.UUIDGenerator;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL        . <br/>
 * The class responsible for workspace storage initialization including:
 * - root node - for all workspaces
 * - /jcr:system - for system workspace
 *    /exo:namespaces
 *    /jcr:nodetypes
 *    /jcr:versionStorage
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: WorkspaceInitializer.java 9750 2006-10-25 11:39:50Z geaz $
 *  */

public class WorkspaceInitializer {
  
  protected static Log log = ExoLogger.getLogger("jcr.WorkspaceInitializer");
  
  //private WorkspaceEntry config;
  private String systemWorkspaceName;
  //private String repositoryName;
  private String workspaceName;
  private DataManager dataManager;
  private String accessControlType; 
  //private NamespaceRegistry nsRegistry;
  private NamespaceDataPersister nsPersister;
  private ExtendedNodeTypeManager ntRegistry;
  private NodeTypeDataPersister ntPersister;
  
  public WorkspaceInitializer(
      WorkspaceEntry config, 
      RepositoryEntry repConfig, 
      DataManager dataManager,
//      NamespaceRegistry nsRegistry,
      NamespaceDataPersister nsPersister,
      ExtendedNodeTypeManager ntRegistry,
      NodeTypeDataPersister ntPersister
      ) {
    this.workspaceName = config.getName();
    this.systemWorkspaceName = repConfig.getSystemWorkspaceName();
    //this.repositoryName = repConfig.getName();
    this.dataManager = dataManager;
    //this.nsRegistry = nsRegistry;
    this.nsPersister = nsPersister;
    this.ntRegistry = ntRegistry;
    this.ntPersister = ntPersister;
    this.accessControlType = repConfig.getAuthenticationPolicy();
  }

  public NodeData initWorkspace(InternalQName rootNodeType)
      throws RepositoryException {
    
    if (isWorkspaceInitialized()) {
      return (NodeData) dataManager.getItemData(Constants.ROOT_PATH);
    }

    // Init root (/) node
    NodeData root = initRootNode(rootNodeType);

    log.debug("Root node for " + workspaceName + " initialized. NodeType: "
        + rootNodeType + " system workspace: " + systemWorkspaceName);

    // Init system workspace
    if (workspaceName.equals(systemWorkspaceName)) {
      // initialize /jcr:system
      NodeData sys = initJcrSystemNode(root);
    }

//    log.info("Workspace " + workspaceName + "@" + this.repositoryName + " is initialized");

    return root;
  }

//  ExtendedNodeTypeManager getNodeTypeManager() {
//    return ntRegistry;
//  }
  
//  NamespaceRegistry getNamespaceRegistry() {
//    return nsRegistry;
//  }
  
  public boolean isWorkspaceInitialized() {
    try {
      return dataManager.getItemData(Constants.ROOT_UUID)==null ? false : true;
    } catch (RepositoryException e) {
      return false;
    }
  }

  private NodeData initRootNode(InternalQName rootNodeType)
      throws RepositoryException {
    
    MutableItemDataChangesLog changesLog = new MutableItemDataChangesLog((SessionImpl)null);
    //List<ItemState> changes = new ArrayList<ItemState>();

    TransientNodeData nodeData = new TransientNodeData(Constants.ROOT_PATH,
        Constants.ROOT_UUID, -1, rootNodeType, new InternalQName[0], 0, null,
        new AccessControlList(true));
    changesLog.add(new ItemState(nodeData, ItemChangeState.ADDED));
    
    try {
      TransientPropertyData primaryType = new TransientPropertyData(
          InternalQPath.makeChildPath(nodeData.getQPath(),
              Constants.PRIMARY_TYPE_URI), UUIDGenerator.generate(), -1,
          PropertyType.NAME, nodeData.getUUID(), false);
      primaryType.setValue(new TransientValueData(rootNodeType));

      nodeData.addProperty(primaryType);
      changesLog.add(new ItemState(primaryType, ItemChangeState.ADDED));

    } catch (IllegalPathException e) {
    }

    dataManager.save(changesLog);
    return nodeData;
  }

  private NodeData initJcrSystemNode(NodeData root) throws RepositoryException {

    MutableItemDataChangesLog changesLog = new MutableItemDataChangesLog((SessionImpl)null);
    
    TransientNodeData sysNodeData = TransientNodeData.createNodeData(root,
        Constants.JCR_SYSTEM, Constants.NT_UNSTRUCTURED);
    
    TransientPropertyData primaryType = TransientPropertyData.createPropertyData(
        sysNodeData, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
    primaryType.setValue(new TransientValueData(sysNodeData.getPrimaryTypeName()));
    
    changesLog.add(ItemState.createAdded(sysNodeData))
              .add(ItemState.createAdded(primaryType));
    
    boolean addACL = !accessControlType.equals(AccessControlPolicy.DISABLE);
    
    if (addACL) {
      AccessControlList acl = new AccessControlList(true); 
      sysNodeData.setMixinTypeNames(new InternalQName[] {Constants.EXO_ACCESS_CONTROLLABLE} );
      TransientPropertyData exoOwner = TransientPropertyData.createPropertyData(
          sysNodeData, Constants.EXO_OWNER, PropertyType.STRING, false, 
          new TransientValueData(acl.getOwner()));
      TransientPropertyData exoPerms = TransientPropertyData.createPropertyData(
          sysNodeData, Constants.EXO_PERMISSIONS, ExtendedPropertyType.PERMISSION, true);
      List <ValueData> perms = new ArrayList <ValueData> ();
      for(int i=0; i<acl.getPermissionEntries().size(); i++) {
        String entry = acl.getPermissionEntries().get(i);
        perms.add(new TransientValueData(entry));
      }
      exoPerms.setValues(perms);
      changesLog.add(ItemState.createAdded(exoOwner))
                .add(ItemState.createAdded(exoPerms));
    }
    
    // init version storage
    TransientNodeData versionStorageNodeData = TransientNodeData
        .createNodeData(sysNodeData, Constants.JCR_VERSIONSTORAGE,
            Constants.EXO_VERSIONSTORAGE);

    TransientPropertyData vsPrimaryType = TransientPropertyData
        .createPropertyData(versionStorageNodeData, Constants.JCR_PRIMARYTYPE,
            PropertyType.NAME, false);
    vsPrimaryType.setValue(new TransientValueData(versionStorageNodeData
        .getPrimaryTypeName()));

    changesLog.add(ItemState.createAdded(versionStorageNodeData)).add(
        ItemState.createAdded(vsPrimaryType));
    
    dataManager.save(changesLog);
    changesLog.clear();

    nsPersister.initStorage(sysNodeData, addACL,
        NamespaceRegistryImpl.DEF_NAMESPACES);
    
    // TODO make one method!
    ntPersister.initNodetypesRoot(sysNodeData, addACL);
    ntPersister.initStorage(((EntityCollection)ntRegistry.getAllNodeTypes()).getList());

    //((NodeTypeManagerImpl) getNodeTypeManager()).initStorage(sysNodeData);

    
    // TODO To catch exceptions caused by already existing nodes (jcr:system, jcr:versionStorage)
    return sysNodeData;
  }
}
