/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.core.access;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;

import javax.jcr.RepositoryException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlEntry;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.access.AccessControlPolicy;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.access.SystemIdentity;
import org.exoplatform.services.jcr.config.RepositoryEntry;
import org.exoplatform.services.jcr.impl.core.CredentialsImpl;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.organization.Group;
import org.exoplatform.services.organization.Membership;
import org.exoplatform.services.organization.OrganizationService;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: AccessManager.java 6678 2006-07-03 11:10:18Z geaz $
 */

public class AccessManager {

  protected static Log log = ExoLogger.getLogger("jcr.RepositoryService");

  private OrganizationService orgService;
  
  final private String accessControlPolicy;

  public AccessManager(RepositoryEntry config, OrganizationService orgService)
      throws RepositoryException {
    this.orgService = orgService;
    this.accessControlPolicy = config.getAccessControl();
  }
  
  /**
   * Policy types: disabled, optional(access control is set on demand), mandatory
   * @return
   */
  public String getAccessControlPolicy() {
    return accessControlPolicy;
  }
  
 public  boolean isDisable(){
    return accessControlPolicy.equals(AccessControlPolicy.DISABLE);
  }
 public boolean isOptional(){
    return accessControlPolicy.equals(AccessControlPolicy.OPTIONAL);
  }
 public boolean isMandatory(){
    return accessControlPolicy.equals(AccessControlPolicy.MANDATORY);
  }
  
  public boolean hasPermission(AccessControlList acl, String permission,
      String userId) throws RepositoryException {
    return hasPermission(acl, parseStringPermissions(permission), userId);
  }

  /**
   * If there are corresponding permission to this node's children. READ - read
   * child items ADD_NODE - add child node SET_PROPERTY - set child property
   * REMOVE - remove child items
   * 
   * @param parent
   * @param permission
   * @return
   * @throws RepositoryException
   */
  public boolean hasPermission(AccessControlList acl, String[] permission,
      String userId) {
    
    if (userId.equals(SystemIdentity.SYSTEM)) {
      // SYSTEM has permission everywhere
      return true;
    } else if (userId.equals(acl.getOwner())) {      
      //Current user is owner of node so has all privileges
      return true;
    } else if (userId.equals(SystemIdentity.ANONIM) 
        && (permission.length > 1 || !permission[0].equals(PermissionType.READ)) ) {
      //Anonim does not have WRITE permission even for ANY
      //System.out.println(">>>userId "+userId+" "+permission[0]);
      return false;
    } else {
      // Check Other with Org service
      for(AccessControlEntry ace: acl.getPermissionEntries()) {
        if (isUserMatch(ace.getIdentity(), userId) && 
            isPermissionMatch(ace.getPermission(), permission))
          return true;
      }
//      for (Iterator i = acl.getPermissionsMap().keySet().iterator(); i.hasNext();) {
//        String identity = (String) i.next();
//        String[] p = (String[]) acl.getPermissionsMap().get(identity);
//        if (isUserMatch(identity, userId) && isPermissionMatch(p, permission))
//          return true;
//      }
    }
    //log.debug("Has Permission == false for "+userId);
    return false;
  }

  public static CredentialsImpl systemCredentials() {
    return new CredentialsImpl(SystemIdentity.SYSTEM, "".toCharArray());
  }

  private static String[] parseStringPermissions(String str) {
    ArrayList permissions = new ArrayList();
    StringTokenizer parser = new StringTokenizer(str, ",");
    while (parser.hasMoreTokens()) {
      String token = parser.nextToken();
      permissions.add(token);
    }
    String[] perms = new String[permissions.size()];
    for (int i = 0; i < perms.length; i++)
      perms[i] = (String) permissions.get(i);
    return perms;
  }

  private boolean isUserMatch(String identity, String userId) {
    if (identity.equals(SystemIdentity.ANY)) //any
      return true;
    if (identity.indexOf(":") == -1) // || identity.equals(SystemIdentity.ANONIM)) 
      // just user
      return identity.equals(userId);
    else {// group
      Iterator groups;
      String membershipName = identity.substring(0, identity.indexOf(":"));
      String groupName = identity.substring(identity.indexOf(":") + 1);
      try {  
        groups = orgService.getGroupHandler().findGroupsOfUser(userId).iterator();
      } catch (Exception e) {
        log.error("AccessManager.isUserMatch() failed " + e);
        return false;
      }
      if ("*".equals(membershipName)) {
        while (groups.hasNext()) {
          Group group = (Group) groups.next();
          log.debug("Check of user "+userId+" membership. Test if "+
              groupName+" == "+group.getId()+" "+groupName.equals(group.getId()));
          if (groupName.equals(group.getId())) 
            return true;
        }
      } else {
        while (groups.hasNext()) {
          Group group = (Group) groups.next();
          try {
            Iterator memberships = 
              orgService.getMembershipHandler().findMembershipsByUserAndGroup(userId, group.getId()).iterator();
            while (memberships.hasNext()) {
              Membership membership = (Membership) memberships.next();
              if (membership.getMembershipType().equals(membershipName))
                return true;
            }
          } catch (Exception e) {
            log.error("AccessManager.isUserMatch() failed " + e);
            return false;
          }
        }
      }
      return false;
    }
  }

  private boolean isPermissionMatch(String existedPermission,
      String[] testPermissions) {
    try {
      //Value[] values = permProperty.getValues();
      for (int i = 0; i < testPermissions.length; i++) {
        if (existedPermission.equals(testPermissions[i]))
          return true;
      }
    } catch (Exception e) {
      log.error("AccessManager.isPermissionMatch() exception " + e);
    }
    return false;
  }

//private boolean isPermissionMatch(String[] existedPermissions,
//String[] testPermissions) {
//try {
////Value[] values = permProperty.getValues();
//for (int i = 0; i < testPermissions.length; i++) {
//  for (int j = 0; j < existedPermissions.length; j++) {
//    if (existedPermissions[j].equals(testPermissions[i]))
//      return true;
//  }
//}
//} catch (Exception e) {
//log.error("AccessManager.isPermissionMatch() exception " + e);
//}
//return false;
//}

}