/**
 * Copyright (c) 2001-2006 eXo Platform SARL, http://www.exoplatform.com/
 *
 * All rights reserved. Please look at license.txt in info directory for
 * more license detail.
 *
 * -------------------------------------------------------------------------
 * $Id: DownloadItem.java 10464 2006-11-21 18:55:33Z rmarins $
 * -------------------------------------------------------------------------
 */
package org.exoplatform.applet.install;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

import org.exoplatform.applet.deploy.Component;
import org.exoplatform.applet.install.event.InstallDispatcher;
import org.exoplatform.applet.install.event.InstallEvent;
import org.exoplatform.applet.logging.Logger;
import org.exoplatform.applet.logging.file.FileLoggerFactory;
import org.exoplatform.applet.security.RestrictedAction;
import org.exoplatform.applet.security.RestrictedActionFactory;

/**
 * @version $Rev: 10464 $ $Date: 2006-11-22 01:55:33 +0700 (Wed, 22 Nov 2006) $
 * @author <a href="mailto:rafael.marins@exoplatform.com">Rafael Marins</a>
 */
public class DownloadItem {

	private class URLChecker implements Runnable {

		private URL location;

		public URLChecker(URL location) {
			super();
			this.location = location;
		}

		public void run() {

			try {
				URLConnection conn = location.openConnection();
				urlValid = true;
				downloadSize = conn.getContentLength();
			} catch (IOException e) {
				urlValid = false;
			}

			urlChecked = true;
		}
	}

	private static final int BUFFER_SIZE = 4096;

	private static final String LOG_SOURCE_CLASS = "DownloadItem";

	private static final RestrictedAction ra = RestrictedActionFactory.getInstance();

	private static final Logger logger = FileLoggerFactory.getInstance().getLogger(DownloadItem.class.getName());

	public static DownloadItem create(Component dependency, URL downloadUrl,
	        String destination, String tempDir) {

		DownloadItem item = new DownloadItem(dependency, downloadUrl,
		        destination, tempDir);
		item.checkUrl();

		return item;
	}

	private Component dependency;

	private URL downloadUrl;

	private String destination;

	private String tempDir;

	private int downloadSize;

	private boolean urlChecked;

	private boolean urlValid;

	private int downloadCount;

	private Thread checkerThread;

	private DownloadItem(Component dependency, URL downloadUrl,
	        String destination, String tempDir) {
		super();
		this.dependency = dependency;
		this.downloadUrl = downloadUrl;
		this.destination = destination;
		this.tempDir = tempDir;
	}

	private void checkUrl() {
		checkerThread = new Thread(new URLChecker(downloadUrl));
		checkerThread.start();
	}

	public synchronized boolean download(InstallDispatcher dispatcher) {

		final String log_method = "download".intern();

		logger.entering(LOG_SOURCE_CLASS, log_method);

		downloadCount = 0;

		InputStream in = null;
		BufferedOutputStream out = null;
		File tempFile = null;

		noteStartDownload(dispatcher);

		try {
			URLConnection conn = downloadUrl.openConnection();
			in = conn.getInputStream();

			// prepare temporary download file
			File tempLocation = new File(tempDir);

			if (!ra.exists(tempLocation)) {
				if (!ra.mkdirs(tempLocation)) {
					throw new IOException("Unable to create directory: "
					        + tempLocation);
				}
			}

			tempFile = File.createTempFile(dependency.getArtifactId() + "-", ".download", tempLocation);

			// download archive to temp location
			out = new BufferedOutputStream(ra.getFileOutputStream(tempFile));

			byte block[] = new byte[BUFFER_SIZE];
			int bytesRead;

			InstallEvent progressEvent = new InstallEvent(this,
			        InstallEvent.DOWNLOAD_PROGRESS);

			while ((bytesRead = in.read(block)) != -1) {
				out.write(block, 0, bytesRead);
				out.flush();
				downloadCount = +bytesRead;

				dispatcher.fire(progressEvent);
			}

			in.close();
			out.close();

			logger.finest("Download of '" + tempFile + "' completed.");

			// prepare final download location
			File destFile = new File(getDestination());

			File destParentDir = destFile.getParentFile();
			if (!ra.exists(destParentDir)) {
				if (!ra.mkdirs(destParentDir)) {
					throw new IOException("Unable to create directory: "
					        + destParentDir);
				}
			}

			copy(tempFile, destFile);
			ra.delete(tempFile);

			noteCompleteDownload(dispatcher);

		} catch (IOException e) {

			logger.throwing(LOG_SOURCE_CLASS, log_method, e);

			if (tempFile != null)
				ra.delete(tempFile);

			noteDownloadFailure(dispatcher);
			return false;

		} finally {
			try {
				if (out != null)
					out.close();
				if (in != null)
					in.close();
			} catch (Exception e) {
				// do nothing
			}
		}

		logger.exiting(LOG_SOURCE_CLASS, log_method);

		return true;
	}

	public Component getDependency() {
		return dependency;
	}

	public String getDestination() {
		return destination;
	}

	public int getDownloadCount() {
		return downloadCount;
	}

	public int getDownloadSize() {
		if (!isUrlChecked()) {
			try {
				checkerThread.join();
			} catch (InterruptedException e) {
				// do nothing
			}
		}
		return downloadSize;
	}

	public URL getDownloadUrl() {
		return downloadUrl;
	}

	public boolean isUrlChecked() {
		return urlChecked;
	}

	public boolean isUrlValid() {
		return urlValid;
	}

	private void noteCompleteDownload(InstallDispatcher d) {
		d.fire(new InstallEvent(this, InstallEvent.DOWNLOAD_COMPLETE));
	}

	private void noteDownloadFailure(InstallDispatcher d) {
		d.fire(new InstallEvent(this, InstallEvent.DOWNLOAD_FAILURE));
	}

	private void noteStartDownload(InstallDispatcher d) {
		d.fire(new InstallEvent(this, InstallEvent.DOWNLOAD_START));
	}

    private static void copy(File src, File dest) throws IOException {

    	FileInputStream inStream = ra.getFileInputStream(src);
    	FileOutputStream outStream = ra.getFileOutputStream(dest);

    	try {
	    	FileChannel sourceChannel = inStream.getChannel();
	    	FileChannel destChannel = outStream.getChannel();

	    	ByteBuffer buffer = ByteBuffer.allocateDirect(32 * 1024);

	    	while (sourceChannel.read(buffer) != -1) {
	    		buffer.flip();

	    		while (buffer.hasRemaining())
	    			destChannel.write(buffer);

	    		buffer.clear();
	    	}

	    	buffer = null;

	    	sourceChannel.close();
	    	destChannel.close();

    	} catch (IOException e) {
    		logger.throwing(LOG_SOURCE_CLASS, "copy", e);
    		throw e;
    	} finally {
    		try {
	        	inStream.close();
	        	outStream.close();
    		} catch (Exception e) {
    			// do nothing
    		}
    	}

    }

}
