/**
 * Copyright (c) 2001-2006 eXo Platform SARL, http://www.exoplatform.com/
 *
 * All rights reserved. Please look at license.txt in info directory for
 * more license detail.
 *
 * -------------------------------------------------------------------------
 * $Id: ExoClassLoader.java 8877 2006-09-14 13:51:24Z rmarins $
 * -------------------------------------------------------------------------
 */
package org.exoplatform.applet.install;

import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLStreamHandlerFactory;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;

/**
 * @version $Rev: 8877 $ $Date: 2006-09-14 20:51:24 +0700 (Thu, 14 Sep 2006) $
 * @author <a href="mailto:rafael.marins@exoplatform.com">Rafael Marins</a>
 */
public class ExoClassLoader extends URLClassLoader {

    /**
     * Standard Java package binary name.
     */
    private static final String JAVA_PACKAGE = "java.";

    /**
     * Standard Java extensions package binary name.
     */
    private static final String JAVAX_PACKAGE = "javax.";

    /**
     * To use or not the parent first class loading delegation model.
     */
    private boolean parentFirst;

    /**
     * Isolate Class Loading from parent.
     */
    private boolean ignoreParent;

    /**
     * To use or not the package intelligence self-first mechanism.
     */
    private boolean usePackageIntelligence;

    /**
     * List of known packages loaded from this class loader.
     */
    private Set packagesLocallyKnown;

    /**
     * Parent Class Loader.
     */
    private ClassLoader parent;

    /**
     * @param urls
     * @param parent
     */
    public ExoClassLoader(URL[] urls, ClassLoader parent) {
        super(urls, parent);
        this.parent = parent;
        initialize();
    }

    /**
     * @param urls
     */
    public ExoClassLoader(URL[] urls) {
        super(urls);
        initialize();
    }

    /**
     * @param urls
     * @param parent
     * @param factory
     */
    public ExoClassLoader(URL[] urls, ClassLoader parent,
            URLStreamHandlerFactory factory) {
        super(urls, parent, factory);
        this.parent = parent;
        initialize();
    }

    // ----------------------------------------------------------------------
    //   Getters and Setters
    // ----------------------------------------------------------------------

    /**
     * @return Returns the useParentFirst.
     */
    public boolean isParentFirst() {
        return parentFirst;
    }

    /**
     * @param use The useParentFirst to set.
     */
    public void setParentFirst(boolean use) {
        this.parentFirst = use;
    }

    /**
     * Returns <code>true</code> if the classloader is set isolated. It means
     * that parent-first delegation model is used only for Java system packages.
     *
     * @return Returns the isolated.
     */
    public boolean isIsolated() {
        return ignoreParent;
    }

    /**
     * @param isolated The ignoreParent to set.
     */
    public void setIsolated(boolean isolated) {
        this.ignoreParent = isolated;
    }

    /**
     * @return Returns the usePackageIntelligence.
     */
    public boolean isUsingPackageIntelligence() {
        return usePackageIntelligence;
    }

    /**
     * @param use The usePackageIntelligence to set.
     */
    public void setUsingPackageIntelligence(boolean use) {
        packagesLocallyKnown = (use ? new HashSet() : null);
        this.usePackageIntelligence = use;
    }

    // ----------------------------------------------------------------------
    //   overriding methods
    // ----------------------------------------------------------------------

    /**
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    protected synchronized Class loadClass(String className, boolean resolve)
            throws ClassNotFoundException {

        Class clazz = null;

        clazz = findLoadedClass(className);

        if (clazz == null) {
            // parent first
            if (useParentFirst(className, true)) {
                try {
                    clazz = parent.loadClass(className);
                }
                catch (ClassNotFoundException cnfe) {
                    clazz = findClass(className);
                }
            }
            // self first
            else {
                try {
                    clazz = findClass(className);
                }
                catch (ClassNotFoundException cnfe) {
                    if (isIsolated()) {
                        throw cnfe;
                    }
                    else {
                        clazz = parent.loadClass(className);
                    }
                }
            }
        }

        if (resolve) {
            resolveClass(clazz);
        }

        return clazz;
    }

    /**
     * @see java.net.URLClassLoader#findClass(java.lang.String)
     */
    protected Class findClass(String className) throws ClassNotFoundException {

        Class clazz = super.findClass(className);

        // add to package to the intelligence mechanism
        String packageName = getPackageName(className);
        addLocalKnownPackage(packageName);

        return clazz;
    }

    /**
     * @see java.lang.ClassLoader#getResource(java.lang.String)
     */
    public URL getResource(String resourceName) {

        URL resource = null;

        // parent first
        if (useParentFirst(resourceName, false)) {
            resource = parent.getResource(resourceName);

            if (resource == null) {
                resource = findResource(resourceName);
            }
        }
        // self first
        else {
            resource = findResource(resourceName);

            if (resource == null && !isIsolated()) {
                resource = parent.getResource(resourceName);
            }
        }

        return resource;
    }

    /**
     * @see java.lang.ClassLoader#getResources(java.lang.String)
     */
    public Enumeration getResources(String resourceName) throws IOException {

        Enumeration resources = null;

        // parent first
        if (useParentFirst(resourceName, false)) {
            resources = parent.getResources(resourceName);

            if (resources == null) {
                resources = findResources(resourceName);
            }
        }
        // self first
        else {
            resources = findResources(resourceName);

            if (resources == null && !isIsolated()) {
                resources = parent.getResources(resourceName);
            }
        }

        return resources;
    }

    // ----------------------------------------------------------------------
    //   implementation specific methods
    // ----------------------------------------------------------------------

    private void initialize() {

        this.parentFirst = true;
        this.ignoreParent = false;
        this.usePackageIntelligence = false;

        // the three attemps to determine the parent classloader

        if (parent == null) {
            parent = Thread.currentThread().getContextClassLoader();
        }

        if (parent == null) {
            parent = ExoClassLoader.class.getClassLoader();
        }

        if (parent == null) {
            parent = getSystemClassLoader();
        }

    }

    /**
     * 
     * @param resourceName fully qualified class named or resource.
     * @param isClass if the checking is used to find a class.
     * <code>false</code> indicates it's locating a general resource, so
     * package intelligence mechanism doesn't work.
     * @return
     */
    private boolean useParentFirst(String resourceName, boolean isClass) {

        // begin with the chosen delegation model
        boolean useParentFirst = isParentFirst() && !isIsolated();

        // try self-first based on intelligent mechanism of known packages 
        if (useParentFirst) {

            if (isUsingPackageIntelligence() && isClass) {

                String packageName = getPackageName(resourceName);

                // self-first <-- when parent first is set to false
                useParentFirst = !knowLocalPackage(packageName);
            }
        }

        // always use parent class loader when loading java packages
        if (!useParentFirst) {
            useParentFirst = (JAVA_PACKAGE.startsWith(resourceName) ||
                    JAVAX_PACKAGE.startsWith(resourceName));
        }

        return useParentFirst;
    }

    private void addLocalKnownPackage(String packageName) {
        if (usePackageIntelligence) {
            packagesLocallyKnown.add(packageName);
        }
    }

    private boolean knowLocalPackage(String packageName) {

        boolean known = false;

        if (packagesLocallyKnown != null) {
            known = packagesLocallyKnown.contains(packageName);
        }

        return known;
    }

    /**
     * Get the complete package name from a fully qualified class name.
     *
     * @param className fully qualified class name. 
     * @return <code>String</code> complete package name.
     */
    private String getPackageName(String className) {

        int lastSeparator = className.lastIndexOf('.');
        String packageName = className.substring(0, lastSeparator);

        return packageName;
    }


}
