/**
 * Copyright (c) 2001-2006 eXo Platform SARL, http://www.exoplatform.com/
 *
 * All rights reserved. Please look at license.txt in info directory for
 * more license detail.
 *
 * -------------------------------------------------------------------------
 * $Id: Installer.java 10466 2006-11-21 19:48:45Z rmarins $
 * -------------------------------------------------------------------------
 */
package org.exoplatform.applet.install;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.exoplatform.applet.deploy.Component;
import org.exoplatform.applet.deploy.LauncherConfiguration;
import org.exoplatform.applet.install.event.InstallDispatcher;
import org.exoplatform.applet.install.event.InstallEvent;
import org.exoplatform.applet.install.event.InstallListener;
import org.exoplatform.applet.logging.Logger;
import org.exoplatform.applet.logging.file.FileLoggerFactory;
import org.exoplatform.applet.security.RestrictedAction;
import org.exoplatform.applet.security.RestrictedActionFactory;

/**
 * @version $Rev: 10466 $ $Date: 2006-11-22 02:48:45 +0700 (Wed, 22 Nov 2006) $
 * @author <a href="mailto:rafael.marins@exoplatform.com">Rafael Marins</a>
 */
public class Installer implements Runnable {

	private static final char DOT = '.';

	private static final char SLASH = '/';

	private static final char JOIN = '-';

	private static final String EXO_BASE_DIR_NAME = ".exo";

	private static final RestrictedAction access = RestrictedActionFactory.getInstance();

	private static final String INSTALLER_THREAD_NAME = "InstallerThread";

	private static final String LOG_SOURCE_CLASS = "Installer";

	private static final Logger logger = FileLoggerFactory.getInstance().getLogger(Installer.class.getName());

	public static URL getDownloadUrl(String baseUrl, String repoBaseUrl,
	        String groupId, String artifactId, String version,
	        String classifier, String type) {

		URL url = null;

		String sSlash = "/".intern();

		StringBuffer location = new StringBuffer(baseUrl);
	
		if (baseUrl.endsWith(sSlash) && repoBaseUrl.endsWith(sSlash))
			location.append(repoBaseUrl.substring(1));
		else if (!baseUrl.endsWith(sSlash) && !repoBaseUrl.endsWith(sSlash))
			location.append(SLASH + repoBaseUrl);
		else
			location.append(repoBaseUrl);

		if (repoBaseUrl.endsWith(sSlash))
			location.deleteCharAt(location.length() - 1);

		location.append(getRepositoryRelativePath(groupId, artifactId, version,
						classifier, type, SLASH));

		try {
	        url = new URL(location.toString());
        } catch (MalformedURLException e) {
        	// do nothing
        }

		return url;
	}

	public static String getInstallPath(String installDirName) {
		return access.getUserHome() + File.separator + EXO_BASE_DIR_NAME
				+ File.separator + installDirName;
	}

	public static String getInstallPath(String installBaseDirName, String scope,
	        String groupId, String artifactId, String version,
	        String classifier, String type) {

		String installPath = getInstallPath(installBaseDirName) + File.separator
				+ scope + getRepositoryRelativePath(groupId, artifactId, version,
						classifier, type, File.separatorChar);

		return installPath;
	}

	public static String getRepositoryRelativePath(String groupId,
	        String artifactId, String version, String classifier,
	        String type, char separator) {

		String pkg = groupId.replace(DOT, separator);

		StringBuffer path = new StringBuffer(separator + pkg + separator
				+ artifactId + separator + version + separator + artifactId
				+ JOIN + version);

		if (classifier != null) {
			path.append(JOIN + classifier);
		}

		path.append(DOT + type.toString());

		return path.toString();
	}

	public static boolean isInstalled(String installBaseDirName, String scope,
	        String groupId, String artifactId, String version,
	        String classifier, String type) {

		String installPath = getInstallPath(installBaseDirName, scope, groupId,
				artifactId, version, classifier, type);

		File f = new File(installPath);
		
		return access.exists(f);
	}

	private String baseUrl;

	private LauncherConfiguration config;

	private Collection itemsToDownload;

	private InstallDispatcher dispatcher;

	public Installer(String baseUrl, LauncherConfiguration cfg) {
		super();
		this.baseUrl = baseUrl;
		this.config = cfg;
	}

	public void addInstallListener(InstallListener listener) {
		getDispatcher().add(listener);
	}

	public Thread createThread() {
		return new Thread(this, INSTALLER_THREAD_NAME);
	}

	private InstallDispatcher getDispatcher() {
		if (dispatcher == null)
			dispatcher = new InstallDispatcher();
		return dispatcher;
	}

	public Collection getDownloadableItems() {
		if (itemsToDownload == null)
			itemsToDownload = prepareNotInstalledComponents();

		return itemsToDownload;
	}

	private void noteBeginInstall() {
		InstallDispatcher d = getDispatcher();
		d.fire(new InstallEvent(this, InstallEvent.INSTALL_BEGIN));
	}

	private void noteEndInstall() {
		InstallDispatcher d = getDispatcher();
		d.fire(new InstallEvent(this, InstallEvent.INSTALL_END));
	}

	private synchronized Collection prepareNotInstalledComponents() {

		Collection pendingComponents = new Vector();

		List dependencies = config.getComponents();

		String tempDir = Installer.getTempDir(config.getInstallBaseDirName());

        Iterator it = dependencies.iterator();
        while (it.hasNext()) {
            Component dep = (Component) it.next();

            if (!Installer.isInstalled(config.getInstallBaseDirName(),
                    dep.getScope(), dep.getGroupId(), dep.getArtifactId(), dep
                            .getVersion(), dep.getClassifier(), dep.getType())) {

                URL downloadUrl = Installer.getDownloadUrl(baseUrl,
                        config.getRepoBaseUrl(), dep.getGroupId(), dep
                                .getArtifactId(), dep.getVersion(), dep
                                .getClassifier(), dep.getType());

                String installLocation = Installer.getInstallPath(config
                        .getInstallBaseDirName(), dep.getScope(), dep
                        .getGroupId(), dep.getArtifactId(), dep.getVersion(),
                        dep.getClassifier(), dep.getType());

                DownloadItem d = DownloadItem.create(dep, downloadUrl,
                        installLocation, tempDir);

                pendingComponents.add(d);
            }
        }

		return pendingComponents;
	}

	/**
     * @param installBaseDirName
     * @return
     */
    private static String getTempDir(String installBaseDirName) {

    	String appDir = getInstallPath(installBaseDirName);
	    return appDir + File.separator + "temp";
    }

	public void run() {

		final String log_method = "run".intern();

		logger.entering(LOG_SOURCE_CLASS, log_method);

		boolean upToDate = true;

		noteBeginInstall();

		// download not installed components to local system
		Iterator it = getDownloadableItems().iterator();
		while (it.hasNext()) {
			DownloadItem di = (DownloadItem) it.next();
			if (!di.download(getDispatcher())) {
				upToDate = false;
				break;
			}
		}

		// call loader to bootstrap components
		if (upToDate) {
			Loader loader = new Loader(config);
			loader.load(getDispatcher());
		} else {
			logger.severe("Download failed. Components are not up to date.");
		}

		noteEndInstall();

		logger.exiting(LOG_SOURCE_CLASS, log_method);
	}

    public long getTotalDownloadSize() {

    	final String log_method = "getTotalDownloadSize".intern();

    	logger.entering(LOG_SOURCE_CLASS, log_method);

    	Collection items = getDownloadableItems();

    	long totalSize = 0;
    	Iterator it = items.iterator();

    	while (it.hasNext()) {

    		DownloadItem d = (DownloadItem) it.next();
    		totalSize += d.getDownloadSize();
    	}

    	logger.exiting(LOG_SOURCE_CLASS, log_method,
    			String.valueOf(totalSize));

	    return totalSize;
    }

}
