package org.esupportail.cas.server.handlers.database;

import java.sql.Connection;
import java.sql.DriverManager;

import org.dom4j.Element;
import org.esupportail.cas.server.util.RedundantHandler;
import org.esupportail.cas.server.util.Server;
import org.esupportail.cas.server.util.log.Log;

/**
 * This abstract class implements a database server class, inherited by
 * BindDatabaseServer and SearchDatabaseServer. 
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 */
abstract class DatabaseServer extends Server {
	
	/**
	 * The name of the class used to access the database through JDBC.
	 */
	private String jdbcDriver;
	
	/**
	 * The JDBC URL of the database.
	 */
	private String jdbcUrl;
	
	/**
	 * Constructor.
	 *
	 * @param handlerDebug debugging mode of the handler
	 * @param handler      the handler the server will be used by
	 * @param serverElement the XML element that declares the server 
	 * @throws Exception Exception
	 */
	protected DatabaseServer(
			final Boolean handlerDebug,
			final RedundantHandler handler,
			final Element serverElement) throws Exception {
		super(handlerDebug, handler, serverElement);
		traceBegin();

		jdbcDriver = getServerSubElementContent(serverElement, "jdbc_driver", true/*needed*/);
		trace("jdbc_driver = " + jdbcDriver);

		checkClass(jdbcDriver);
		
		jdbcUrl = getServerSubElementContent(serverElement, "jdbc_url", true/*needed*/);
		trace("jdbc_url = " + jdbcUrl);

		traceEnd();
	}
	
	/**
	 * Connect to a database using specified username and password.
	 *
	 * @param bindUsername the username to use for the connection
	 * @param bindPassword the associated password
	 *
	 * @return a Connection object on success, null on error (in this case,
	 * the error code can be retrieved with the connectError() method).
	 */	
	protected final Connection connect(final String bindUsername,
									   final String bindPassword) {
		traceBegin();
		Connection connection;
		try {	
			Class.forName(jdbcDriver);
			trace("Connecting to the database (`" + jdbcUrl + "')...");
			connection = DriverManager.getConnection(jdbcUrl, bindUsername, bindPassword);
		} catch (java.lang.ClassNotFoundException e) {
			trace("Connection failed: " + e.toString());
			Log.warn("Database provider (" + jdbcDriver + ") is probably not installed");
			setConnectError(CONNECT_NOAUTH);
			traceEnd(null);
			return null;
		} catch (java.sql.SQLException e) {
			if (e.getMessage().equals("Server configuration denies access to data source")) {
				trace(e.getMessage());
				setConnectError(CONNECT_NOAUTH);
				traceEnd(null);
				return null;
			} else {
				trace("Connection failed: " + e.getMessage());
				setConnectError(CONNECT_FAILURE);
				traceEnd(null);
				return null;
			}
		} catch (Exception e) {
			trace("Connection failed (" + e.getClass().getName() + "): " + e.getMessage());
			setConnectError(CONNECT_FAILURE);
			traceEnd(null);
			return null;
		} 
		setConnectError(CONNECT_SUCCESS);
		traceEnd("ok");
		return connection;
	}
	
}

